# -*- coding: utf-8 -*-
"takes care of the option's values and multi values"
# Copyright (C) 2013-2017 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from time import time
import sys
import weakref
from .error import ConfigError, SlaveError, PropertiesOptionError
from .setting import owners, expires_time, undefined, forbidden_owners
from .autolib import carry_out_calculation
from .i18n import _
from .option import DynSymLinkOption, Option


class Values(object):
    """The `Config`'s root is indeed  in charge of the `Option()`'s values,
    but the values are physicaly located here, in `Values`, wich is also
    responsible of a caching utility.
    """
    __slots__ = ('context', '_p_', '__weakref__')

    def __init__(self, context, storage):
        """
        Initializes the values's dict.

        :param context: the context is the home config's values

        """
        self.context = weakref.ref(context)
        # the storage type is dictionary or sqlite3
        self._p_ = storage

    #______________________________________________________________________
    # get context

    def _getcontext(self):
        """context could be None, we need to test it
        context is None only if all reference to `Config` object is deleted
        (for example we delete a `Config` and we manipulate a reference to
        old `SubConfig`, `Values`, `Multi` or `Settings`)
        """
        context = self.context()
        if context is None:
            raise ConfigError(_('the context does not exist anymore'))
        return context

    #______________________________________________________________________
    # get value

    def get_cached_value(self,
                         opt,
                         path=None,
                         validate=True,
                         force_permissive=False,
                         trusted_cached_properties=True,
                         setting_properties=undefined,
                         self_properties=undefined,
                         index=None,
                         display_warnings=True):
        context = self._getcontext()
        settings = context.cfgimpl_get_settings()
        if path is None:
            path = opt.impl_getpath(context)
        ntime = None
        if self_properties is undefined:
            self_properties = settings.getproperties(opt,
                                                     path,
                                                     setting_properties=setting_properties,
                                                     index=index)
        if 'cache' in setting_properties and self._p_.hascache(path, index):
            if 'expire' in setting_properties:
                ntime = int(time())
            is_cached, value = self._p_.getcache(path,
                                                 ntime,
                                                 None)
            if index:
                value = value[index]
            if is_cached:
                return value
        val = self.get_validated_value(opt,
                                       path,
                                       validate,
                                       setting_properties,
                                       self_properties,
                                       index=index,
                                       display_warnings=display_warnings,
                                       force_permissive=force_permissive)
        if index is None and 'cache' in setting_properties and \
                validate and force_permissive is False \
                and trusted_cached_properties is True:
            if 'expire' in setting_properties:
                if ntime is None:
                    ntime = int(time())
                ntime = ntime + expires_time
            self._p_.setcache(path, val, ntime, None)
        return val

    def get_validated_value(self,
                            opt,
                            path,
                            validate,
                            setting_properties,
                            self_properties=None,
                            index=None,
                            display_warnings=True,
                            force_permissive=False):
        """same has getitem but don't touch the cache
        index is None for slave value, if value returned is not a list, just return []
        """
        context = self._getcontext()
        config_error = None
        try:
            value = self.getvalue(opt,
                                  path,
                                  index,
                                  setting_properties,
                                  self_properties,
                                  validate,
                                  force_permissive=force_permissive)
        except ConfigError as value:
            value_error = True
            # For calculating properties, we need value (ie for mandatory
            # value).
            # If value is calculating with a PropertiesOptionError's option
            # getvalue raise a ConfigError.
            # We can not raise ConfigError if this option should raise
            # PropertiesOptionError too. So we get config_error and raise
            # ConfigError if properties did not raise.
            config_error = value
            # value is not set, for 'undefined' (cannot set None because of
            # mandatory property)
            value = undefined
        else:
            value_error = False
            if validate:
                err = opt.impl_validate(value,
                                        context,
                                        'validator' in setting_properties,
                                        force_index=index,
                                        display_error=True,
                                        display_warnings=False,
                                        setting_properties=setting_properties)
                if err:
                    config_error = err
                    value = None
        if not value_error and validate and display_warnings:
            opt.impl_validate(value,
                              context,
                              'validator' in setting_properties,
                              force_index=index,
                              display_error=False,
                              display_warnings=display_warnings,
                              setting_properties=setting_properties)
        if config_error is not None:
            raise config_error
        return value

    def getvalue(self,
                 opt,
                 path,
                 index,
                 setting_properties,
                 self_properties,
                 validate,
                 force_permissive=False):
        """actually retrieves the value

        :param opt: the `option.Option()` object
        :returns: the option's value (or the default value if not set)
        """
        force_default = 'frozen' in self_properties and \
            'force_default_on_freeze' in self_properties
        # not default value
        if index is None or not opt.impl_is_master_slaves('slave'):
            _index = None
        else:
            _index = index
        owner, value = self._p_.getowner(path,
                                         owners.default,
                                         index=_index,
                                         with_value=True)
        is_default = owner == owners.default
        if not is_default and not force_default:
            if index is not None and not opt.impl_is_master_slaves('slave'):
                if len(value) > index:
                    return value[index]
                #value is smaller than expected
                #so return default value
            else:
                return value
        return self._getdefaultvalue(opt,
                                     path,
                                     index,
                                     validate,
                                     setting_properties,
                                     force_permissive=force_permissive)

    def getdefaultvalue(self,
                        opt,
                        path,
                        setting_properties=undefined,
                        index=None):
        """get default value:
        - get meta config value or
        - get calculated value or
        - get default value
        :param opt: the `option.Option()` object
        :param path: path for `option.Option()` object
        :type path: str
        :param index: index of a multi/submulti
        :type index: int
        :returns: default value
        """
        return self._getdefaultvalue(opt,
                                     path,
                                     index,
                                     True,
                                     setting_properties)

    def _getdefaultvalue(self,
                         opt,
                         path,
                         index,
                         validate,
                         setting_properties,
                         force_permissive=False):
        context = self._getcontext()
        def _reset_cache():
            # calculated value could be a new value, so reset cache
            context.cfgimpl_reset_cache(opt=opt,
                                        path=path)

        if self._is_meta(opt,
                         path,
                         index,
                         setting_properties,
                         force_permissive=force_permissive):
            meta = context.cfgimpl_get_meta()
            # retrieved value from meta config
            try:
                value = meta.getattr(path,
                                     index=index,
                                     setting_properties=setting_properties,
                                     force_permissive=force_permissive)
            except PropertiesOptionError:
                # if properties error, return an other default value
                # unexpected error, should not happened
                pass
            else:
                return value

        if opt.impl_has_callback():
            # if value has callback, calculate value
            callback, callback_params = opt.impl_get_callback()
            value = carry_out_calculation(opt,
                                          context=context,
                                          callback=callback,
                                          callback_params=callback_params,
                                          setting_properties=setting_properties,
                                          index=index,
                                          validate=validate)
            if isinstance(value, list) and index is not None:
                # if value is a list and index is set
                if opt.impl_is_submulti() and (value == [] or not isinstance(value[0], list)):
                    # return value only if it's a submulti and not a list of list
                    _reset_cache()
                    return value

                if len(value) > index:
                    # return the value for specified index if found
                    _reset_cache()
                    return value[index]
                # there is no calculate value for this index,
                # so return an other default value
            elif isinstance(value, list):
                # value is a list, but no index specified
                _reset_cache()
                if opt.impl_is_submulti() and (value == [] or not isinstance(value[0], list)):
                    # if submulti, return a list of value
                    return [value]
                # otherwise just return the value
                return value
            elif index is not None:
                # if not list but with index
                _reset_cache()
                if opt.impl_is_submulti():
                    # if submulti, return a list of value
                    return [value]
                # otherwise just return the value
                return value
            else:
                _reset_cache()
                # not a list or index is None
                if opt.impl_is_submulti():
                    # return a list of list for a submulti
                    return [[value]]
                elif opt.impl_is_multi():
                    # return a list for a multi
                    return [value]
                # not a list, return value
                return value

        # now try to get default value:
        # - if opt is a submulti, return a list a list
        # - if opt is a multi, return a list
        # - default value
        value = opt.impl_getdefault()
        if opt.impl_is_multi() and index is not None:
            # if index, must return good value for this index
            if len(value) > index:
                value = value[index]
            else:
                # no value for this index, retrieve default multi value
                # default_multi is already a list for submulti
                value = opt.impl_getdefault_multi()
        return value

    def isempty(self,
                opt,
                value,
                force_allow_empty_list=False,
                index=None):
        "convenience method to know if an option is empty"
        if value is undefined:
            return False
        else:
            empty = opt._empty
            if index in [None, undefined] and opt.impl_is_multi():
                if force_allow_empty_list:
                    allow_empty_list = True
                else:
                    allow_empty_list = opt.impl_allow_empty_list()
                    if allow_empty_list is undefined:
                        if opt.impl_is_master_slaves('slave'):
                            allow_empty_list = True
                        else:
                            allow_empty_list = False
                isempty = value is None or (not allow_empty_list and value == []) or \
                    None in value or empty in value
            else:
                isempty = value is None or value == empty
        return isempty

    def get_modified_values(self):
        return self._p_.get_modified_values()

    #______________________________________________________________________
    # set value

    def setvalue(self,
                 opt,
                 value,
                 path,
                 force_permissive,
                 index,
                 setting_properties,
                 _commit):

        context = self._getcontext()
        owner = context.cfgimpl_get_settings().getowner()
        if 'validator' in setting_properties:
            if opt._has_consistencies():
                # set value to a fake config when option has dependency
                # validation will be complet in this case (consistency, ...)
                tested_context = context._gen_fake_values()
                tested_values = tested_context.cfgimpl_get_values()
                tested_values._setvalue(opt,
                                        path,
                                        value,
                                        index=index,
                                        owner=owner)
            else:
                tested_context = context
                tested_values = self
            tested_values.validate_setitem(opt,
                                           value,
                                           path,
                                           force_permissive,
                                           setting_properties,
                                           index)

        self._setvalue(opt,
                       path,
                       value,
                       owner,
                       index=index,
                       commit=_commit)

    def validate_setitem(self,
                         opt,
                         value,
                         path,
                         force_permissive,
                         setting_properties,
                         index):

        context = self._getcontext()
        settings = context.cfgimpl_get_settings()
        # First validate properties with this value
        properties = settings.getproperties(opt,
                                            path,
                                            setting_properties=setting_properties,
                                            index=index)
        if 'everything_frozen' in setting_properties or 'frozen' in properties:
            datas = {'opt': opt,
                     'path': path,
                     'setting_properties': setting_properties,
                     'index': index,
                     'debug': True}
            raise PropertiesOptionError(None,
                                        ['frozen'],
                                        settings,
                                        datas,
                                        'option')
        if settings.validate_mandatory(opt,
                                       index,
                                       value,
                                       setting_properties,
                                       properties):
            datas = {'opt': opt,
                     'path': path,
                     'setting_properties': setting_properties,
                     'index': index,
                     'debug': True}
            raise PropertiesOptionError(None,
                                        ['mandatory'],
                                        settings,
                                        datas,
                                        'option')
        # Value must be valid for option
        opt.impl_validate(value,
                          context,
                          display_warnings=False,
                          force_index=index,
                          setting_properties=setting_properties)
        # No error found so emit warnings
        opt.impl_validate(value,
                          context,
                          display_error=False,
                          force_index=index,
                          setting_properties=setting_properties)

    def _setvalue(self,
                  opt,
                  path,
                  value,
                  owner,
                  index=None,
                  commit=True):

        self._getcontext().cfgimpl_reset_cache(opt=opt,
                                               path=path)
        if isinstance(value, list):
            # copy
            value = list(value)
        self._p_.setvalue(path,
                          value,
                          owner,
                          index,
                          commit)

    def _is_meta(self,
                 opt,
                 path,
                 index,
                 setting_properties,
                 force_permissive=False,
                 force_owner_is_default=False):

        if not force_owner_is_default and self._p_.hasvalue(path,
                                                            index=index):
            # has already a value, so not meta
            return False
        context = self._getcontext()
        meta = context.cfgimpl_get_meta()
        if meta is None:
            return False
        if opt.impl_is_master_slaves('slave'):
            master = opt.impl_get_master_slaves().getmaster()
            masterp = master.impl_getpath(context)
            # slave could be a "meta" only if master hasn't value
            if self._p_.hasvalue(masterp,
                                 index=index):
                return False
        return not meta.cfgimpl_get_values().is_default_owner(opt,
                                                              path,
                                                              setting_properties,
                                                              index=index,
                                                              force_permissive=force_permissive)


    #______________________________________________________________________
    # owner

    def getowner(self,
                 opt,
                 path,
                 setting_properties,
                 index=None,
                 force_permissive=False):
        """
        retrieves the option's owner

        :param opt: the `option.Option` object
        :param force_permissive: behaves as if the permissive property
                                 was present
        :returns: a `setting.owners.Owner` object
        """
        if opt.impl_is_symlinkoption() and \
                not isinstance(opt, DynSymLinkOption):
            opt = opt.impl_getopt()
        return self._getowner(opt,
                              path,
                              setting_properties,
                              index=index,
                              force_permissive=force_permissive)

    def _getowner(self,
                  opt,
                  path,
                  setting_properties,
                  force_permissive=False,
                  validate_meta=undefined,
                  self_properties=undefined,
                  only_default=False,
                  index=None):
        """get owner of an option
        """
        #FIXME: validate_meta ne marche que si == False ou undefined !
        if validate_meta is not False and validate_meta is not undefined:
            raise Exception('poeut')
        #if not isinstance(opt, Option) and not isinstance(opt,
        #                                                  DynSymLinkOption):
        #    raise ConfigError(_('owner only avalaible for an option'))
        #FIXME pas defaut fait ??
        context = self._getcontext()
        if self_properties is undefined:
            self_properties = context.cfgimpl_get_settings().getproperties(opt,
                                                                           path,
                                                                           setting_properties)
        if 'frozen' in self_properties and 'force_default_on_freeze' in self_properties:
            return owners.default
        if only_default:
            if self._p_.hasvalue(path,
                                 index):
                owner = undefined
            else:
                owner = owners.default
        else:
            owner = self._p_.getowner(path,
                                      owners.default,
                                      index=index)
        if owner is owners.default and validate_meta is not False:
            if validate_meta is undefined:
                validate_meta = self._is_meta(opt,
                                              path,
                                              index,
                                              setting_properties,
                                              force_permissive=force_permissive,
                                              force_owner_is_default=True)
            if validate_meta:
                owner = owners.meta
        return owner

    def setowner(self,
                 path,
                 owner,
                 index=None):
        """
        sets a owner to an option

        :param opt: the `option.Option` object
        :param owner: a valid owner, that is a `setting.owners.Owner` object
        """
        if not isinstance(owner, owners.Owner):
            raise TypeError(_("invalid owner {0}").format(str(owner)))

        if owner in forbidden_owners:
            raise ConfigError(_('set owner "{0}" is forbidden').format(str(owner)))

        if not self._p_.hasvalue(path):
            raise ConfigError(_('no value for {0} cannot change owner to {1}'
                                '').format(path, owner))
        self._p_.setowner(path, owner, index=index)

    def is_default_owner(self,
                         opt,
                         path,
                         setting_properties,
                         validate_meta=undefined,
                         self_properties=undefined,
                         index=None,
                         force_permissive=False):
        owner = self._getowner(opt,
                               path,
                               setting_properties,
                               validate_meta=validate_meta,
                               self_properties=self_properties,
                               only_default=True,
                               index=index,
                               force_permissive=force_permissive)
        return owner == owners.default

    #______________________________________________________________________
    # reset

    def reset(self,
              opt,
              path,
              setting_properties,
              validate=True,
              _commit=True,
              force_permissive=False):

        context = self._getcontext()
        setting = context.cfgimpl_get_settings()
        hasvalue = self._p_.hasvalue(path)

        if validate and hasvalue and 'validator' in setting_properties:
            fake_context = context._gen_fake_values()
            fake_value = fake_context.cfgimpl_get_values()
            fake_value.reset(opt,
                             path,
                             setting_properties,
                             validate=False)
            #FIXME ce n'est pas deja fait dans le reset ?
            #sinon c'est un validate_properties qu'il faut faire ...
            #fake_value.get_cached_value(opt,
            #                            path,
            #                            setting_properties=setting_properties,
            #                            force_permissive=force_permissive)
        if opt.impl_is_master_slaves('master'):
            opt.impl_get_master_slaves().reset(opt,
                                               self,
                                               setting_properties,
                                               _commit=_commit,
                                               force_permissive=force_permissive)
        if hasvalue:
            if 'force_store_value' in setting.getproperties(opt,
                                                            path,
                                                            setting_properties,
                                                            apply_requires=False):
                value = self._getdefaultvalue(opt,
                                              path,
                                              None,
                                              validate,
                                              setting_properties)
                self._setvalue(opt,
                               path,
                               value,
                               owners.forced,
                               None,
                               commit=_commit)
            else:
                self._p_.resetvalue(path,
                                    _commit)
            context.cfgimpl_reset_cache(opt=opt,
                                        path=path)

    def reset_slave(self,
                    opt,
                    path,
                    index,
                    setting_properties,
                    validate=True,
                    force_permissive=False):

        context = self._getcontext()
        if validate and 'validator' in setting_properties:
            fake_context = context._gen_fake_values()
            fake_value = fake_context.cfgimpl_get_values()
            fake_value.reset_slave(opt,
                                   path,
                                   index,
                                   setting_properties,
                                   validate=False)
            fake_value.get_cached_value(opt,
                                        path,
                                        index=index,
                                        setting_properties=setting_properties,
                                        force_permissive=force_permissive)
        self._p_.resetvalue_index(path, index)

    def reset_master(self,
                     subconfig,
                     opt,
                     path,
                     index,
                     force_permissive,
                     setting_properties):

        current_value = self.get_cached_value(opt,
                                              path,
                                              setting_properties=setting_properties,
                                              force_permissive=force_permissive)
        current_value.pop(index)
        self.setvalue(opt,
                      current_value,
                      path,
                      force_permissive=force_permissive,
                      index=None,
                      setting_properties=setting_properties,
                      _commit=True)
        subconfig.cfgimpl_get_description().pop(self,
                                                index,
                                                setting_properties,
                                                force_permissive)

    #______________________________________________________________________
    # information

    def set_information(self, key, value):
        """updates the information's attribute

        :param key: information's key (ex: "help", "doc"
        :param value: information's value (ex: "the help string")
        """
        self._p_.set_information(key, value)

    def get_information(self, key, default=undefined):
        """retrieves one information's item

        :param key: the item string (ex: "help")
        """
        return self._p_.get_information(key, default)

    def del_information(self, key, raises=True):
        self._p_.del_information(key, raises)

    #______________________________________________________________________
    # mandatory warnings

    def mandatory_warnings(self, force_permissive=True):
        """convenience function to trace Options that are mandatory and
        where no value has been set

        :returns: generator of mandatory Option's path
        """
        context = self._getcontext()
        settings = context.cfgimpl_get_settings()
        setting_properties = context.cfgimpl_get_settings().get_context_properties()
        setting_properties.update(['mandatory', 'empty'])
        def _is_properties_option(err, path):
            #FIXME hum ...
            if not isinstance(err, Exception):
                pass
            elif isinstance(err, PropertiesOptionError):
                if err.proptype == ['mandatory']:
                    return path
            elif isinstance(err, ConfigError):
                #assume that uncalculated value is an empty value
                return path
            else:
                raise err

        def _mandatory_warnings(description, currpath=None):
            if currpath is None:
                currpath = []
            for opt in description._impl_getchildren(context=context):
                name = opt.impl_getname()
                path = '.'.join(currpath + [name])

                if opt.impl_is_optiondescription():
                    #FIXME ?
                    if not settings.validate_properties(opt, True, False, path=path,
                                                        force_permissive=True,
                                                        setting_properties=setting_properties):
                        for path in _mandatory_warnings(opt, currpath + [name]):
                            yield path
                else:
                    if opt.impl_is_symlinkoption() and \
                            not isinstance(opt, DynSymLinkOption):
                        continue
                    self_properties = settings.getproperties(opt,
                                                             path,
                                                             setting_properties=setting_properties)
                    if 'mandatory' in self_properties or 'empty' in self_properties:
                        err = self.get_cached_value(opt, path=path,
                                                    trusted_cached_properties=False,
                                                    force_permissive=True,
                                                    setting_properties=setting_properties,
                                                    self_properties=self_properties,
                                                    validate=True,
                                                    display_warnings=False)
                        if opt.impl_is_master_slaves('slave') and isinstance(err, list):
                            for val in err:
                                ret = _is_properties_option(val, path)
                                if ret is not None:
                                    yield ret
                                    break
                        else:
                            ret = _is_properties_option(err, path)
                            if ret is not None:
                                yield ret

        descr = context.cfgimpl_get_description()
        for path in _mandatory_warnings(descr):
            yield path
